/*
 * QLogic Fibre Channel HBA Driver
 * Copyright (c)  2003-2005 QLogic Corporation
 *
 * See LICENSE.qla2xxx for copyright and licensing details.
 */
#include "qla_def.h"

#include <linux/debugfs.h>
#include <linux/vmalloc.h>

static struct dentry *qla2x00_dfs_root;
static atomic_t qla2x00_dfs_root_count;

/* DISTRO/KERNEL compatibility.
 *
 * Unfortunately, Not all distro variants which have a 2.6.18 base share the
 * same 'struct inode' infrastructure.  OpenSuSE 10.2 maintains the old
 * 'u.generic_ip' members, whereas RHEL5 has migrated over to the 'i_private'
 * usages.  We're going to be RHEL5 biased to handle a larger base of users.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#define get_inode_private(n)	((n)->u.generic_ip)
#else
#define get_inode_private(n)	((n)->i_private)
#endif

#define fce_calc_vbuf_size(n)	((n) * (32 * 100))
#define fappend(fmt, x...)	pos += snprintf(buf + pos, len - pos, fmt , ##x)

static int
qla2x00_dfs_fce_open(struct inode *inode, struct file *file)
{
	scsi_qla_host_t *ha = get_inode_private(inode);
	int rval;

	if (!ha->flags.fce_enabled)
		return 0;

	down(&ha->fce_mutex);

	/* Pause tracing to flush FCE buffers. */
	rval = qla2x00_disable_fce_trace(ha, &ha->fce_wr, &ha->fce_rd);
	if (rval)
		qla_printk(KERN_WARNING, ha,
		    "DebugFS: Unable to disable FCE (%d).\n", rval);

	ha->flags.fce_enabled = 0;

	up(&ha->fce_mutex);

	file->private_data = get_inode_private(inode);

	return 0;
}

static ssize_t
qla2x00_dfs_fce_read(struct file *file, char __user *userbuf, size_t count,
    loff_t *ppos)
{
	scsi_qla_host_t *ha = file->private_data;
	char *buf = ha->fce_vbuf;
	const size_t len = fce_calc_vbuf_size(ha->fce_bufs);
	size_t pos = 0;
	ssize_t res;
	uint32_t cnt;
	uint32_t *fce;
	uint64_t fce_start;

	down(&ha->fce_mutex);

	fappend("FCE Trace Buffer\n");
	fappend("In Pointer = %llx\n\n", (unsigned long long)ha->fce_wr);
	fappend("Base = %llx\n\n", (unsigned long long) ha->fce_dma);
	fappend("FCE Enable Registers\n");
	fappend("%08x %08x %08x %08x %08x %08x\n",
	    ha->fce_mb[0], ha->fce_mb[2], ha->fce_mb[3], ha->fce_mb[4],
	    ha->fce_mb[5], ha->fce_mb[6]);

	fce = (uint32_t *) ha->fce;
	fce_start = (unsigned long long) ha->fce_dma;
	for (cnt = 0; cnt < fce_calc_size(ha->fce_bufs) / 4; cnt++) {
		if (cnt % 8 == 0)
			fappend("\n%llx: ",
			    (unsigned long long)((cnt * 4) + fce_start));
		else
			fappend(" ");
		fappend("%08x", *fce++);
	}
	fappend("\nEnd\n");

	res = simple_read_from_buffer(userbuf, count, ppos, buf, pos);

	up(&ha->fce_mutex);

	return res;
}

static int
qla2x00_dfs_fce_release(struct inode *inode, struct file *file)
{
	scsi_qla_host_t *ha = get_inode_private(inode);
	int rval;

	if (ha->flags.fce_enabled)
		return 0;

	down(&ha->fce_mutex);

	/* Re-enable FCE tracing. */
	ha->flags.fce_enabled = 1;
	memset(ha->fce, 0, fce_calc_size(ha->fce_dbufs));
	rval = qla2x00_enable_fce_trace(ha, ha->fce_dma, ha->fce_dbufs,
	    ha->fce_mb, &ha->fce_bufs);
	if (rval) {
		qla_printk(KERN_WARNING, ha,
		    "DebugFS: Unable to reinitialize FCE (%d).\n", rval);
		ha->flags.fce_enabled = 0;
	}

	up(&ha->fce_mutex);

	return 0;
}

static struct file_operations dfs_fce_ops = {
	.open		= qla2x00_dfs_fce_open,
	.read		= qla2x00_dfs_fce_read,
	.release	= qla2x00_dfs_fce_release,
};

int
qla2x00_dfs_setup(scsi_qla_host_t *ha)
{
	if (!IS_QLA25XX(ha))
		goto out;
	if (!ha->fce)
		goto out;

	if (!ha->fce_vbuf)
		ha->fce_vbuf = vmalloc(fce_calc_vbuf_size(ha->fce_bufs));
	if (!ha->fce_vbuf)
		goto out;

	if (qla2x00_dfs_root)
		goto create_dir;

	atomic_set(&qla2x00_dfs_root_count, 0);
	qla2x00_dfs_root = debugfs_create_dir(QLA2XXX_DRIVER_NAME, NULL);
	if (!qla2x00_dfs_root) {
		qla_printk(KERN_NOTICE, ha,
		    "DebugFS: Unable to create root directory.\n");
		goto out;
	}

create_dir:
	if (ha->dfs_dir)
		goto create_nodes;

	init_MUTEX(&ha->fce_mutex);
	ha->dfs_dir = debugfs_create_dir(ha->host_str, qla2x00_dfs_root);
	if (!ha->dfs_dir) {
		qla_printk(KERN_NOTICE, ha,
		    "DebugFS: Unable to create ha directory.\n");
		goto out;
	}

	atomic_inc(&qla2x00_dfs_root_count);

create_nodes:
	ha->dfs_fce = debugfs_create_file("fce", S_IRUSR, ha->dfs_dir, ha,
	    &dfs_fce_ops);
	if (!ha->dfs_fce) {
		qla_printk(KERN_NOTICE, ha,
		    "DebugFS: Unable to fce node.\n");
		goto out;
	}
out:
	return 0;
}

int
qla2x00_dfs_remove(scsi_qla_host_t *ha)
{
	if (!IS_QLA25XX(ha))
		return 0;

	if (ha->dfs_fce) {
		debugfs_remove(ha->dfs_fce);
		ha->dfs_fce = NULL;
	}

	if (ha->dfs_dir) {
		debugfs_remove(ha->dfs_dir);
		ha->dfs_dir = NULL;
		atomic_dec(&qla2x00_dfs_root_count);
	}

	if (atomic_read(&qla2x00_dfs_root_count) == 0 &&
	    qla2x00_dfs_root) {
		debugfs_remove(qla2x00_dfs_root);
		qla2x00_dfs_root = NULL;
	}

	if (ha->fce_vbuf)
		vfree(ha->fce_vbuf);
	ha->fce_vbuf = NULL;

	return 0;
}
